from hpotk.model import Versioned
import warnings

# TODO[v0.3.0] - remove the deprecated file
warnings.warn('hpotk.ontology._attrs has been deprecated. Use hpotk.model.Versioned instead',
              DeprecationWarning, stacklevel=2)

